@@ -4,6 +4,7 @@
     SPDX-FileCopyrightText: 2000 Matthias Elter <elter@kde.org>
     SPDX-FileCopyrightText: 2003, 2004, 2006 Matthias Kretz <kretz@kde.org>
     SPDX-FileCopyrightText: 2004 Frans Englich <frans.englich@telia.com>
+    SPDX-FileCopyrightText: 2021 Alexander Lohnau <alexander.lohnau@gmx.de>
 
     SPDX-License-Identifier: LGPL-2.0-only
 */
@@ -17,11 +18,12 @@
 #include <QLibrary>
 #include <QVBoxLayout>
 
-#include <KPluginInfo>
-#include <KPluginLoader>
+#include <KAboutData>
+#include <KAuthorized>
 #include <KLocalizedString>
 #include <KMessageBox>
-#include <KAboutData>
+#include <KPluginInfo>
+#include <KPluginLoader>
 
 #include <KQuickAddons/ConfigModule>
 
@@ -64,13 +66,16 @@ KCModule *KCModuleLoader::loadModule(const KCModuleInfo &mod, ErrorReporting rep
 
     if (!mod.isValid()) {
         return reportError(report,
-                           i18n("The module %1 could not be found.",
-                                mod.moduleName()), i18n("<qt><p>The diagnosis is:<br />The desktop file %1 could not be found.</p></qt>", mod.fileName()), parent);
+                           i18n("The module %1 could not be found.", mod.moduleName()),
+                           i18n("<qt><p>The diagnosis is:<br />The desktop file %1 could not be found.</p></qt>", mod.fileName()),
+                           parent);
     }
     if (mod.service() && mod.service()->noDisplay()) {
-        return reportError(report, i18n("The module %1 is disabled.", mod.moduleName()),
-                           i18n("<qt><p>Either the hardware/software the module configures is not available or the module has been disabled by the administrator.</p></qt>"),
-                           parent);
+        return reportError(
+            report,
+            i18n("The module %1 is disabled.", mod.moduleName()),
+            i18n("<qt><p>Either the hardware/software the module configures is not available or the module has been disabled by the administrator.</p></qt>"),
+            parent);
     }
 
     if (!mod.library().isEmpty()) {
@@ -86,10 +91,11 @@ KCModule *KCModuleLoader::loadModule(const KCModuleInfo &mod, ErrorReporting rep
         KCModule *module = nullptr;
 
         KPluginLoader loader(KPluginLoader::findPlugin(QLatin1String("kcms/") + mod.library()));
-        KPluginFactory* factory = loader.factory();
+        KPluginFactory *factory = loader.factory();
         if (!factory) {
             // KF6 TODO: make this a warning, and remove mention of fallback
-            qCDebug(KCMUTILS_LOG) << "Couldn't load plugin" << QLatin1String("kcms/") + mod.library() << ":" << loader.errorString() << " -- falling back to old-style loading from desktop file";
+            qCDebug(KCMUTILS_LOG) << "Couldn't load plugin" << QLatin1String("kcms/") + mod.library() << ":" << loader.errorString()
+                                  << " -- falling back to old-style loading from desktop file";
         } else {
             std::unique_ptr<KQuickAddons::ConfigModule> cm(factory->create<KQuickAddons::ConfigModule>(nullptr, args2));
             if (!cm) {
@@ -110,7 +116,7 @@ KCModule *KCModuleLoader::loadModule(const KCModuleInfo &mod, ErrorReporting rep
         if (module) {
             return module;
         } else
-//#ifndef NDEBUG
+#if KCMUTILS_BUILD_DEPRECATED_SINCE(5, 85)
         {
             // KF6 TODO: remove this old compat block
             // get the create_ function
@@ -123,12 +129,13 @@ KCModule *KCModuleLoader::loadModule(const KCModuleInfo &mod, ErrorReporting rep
                 if (create) {
                     return create(parent, mod.handle().toLatin1().constData());
                 } else {
-                    qCWarning(KCMUTILS_LOG) << "This module has no valid entry symbol at all. The reason could be that it's still using K_EXPORT_COMPONENT_FACTORY with a custom X-KDE-FactoryName which is not supported anymore";
+                    qCWarning(KCMUTILS_LOG) << "This module has no valid entry symbol at all. The reason could be that it's still using "
+                                               "K_EXPORT_COMPONENT_FACTORY with a custom X-KDE-FactoryName which is not supported anymore";
                 }
                 lib.unload();
             }
         }
-//#endif // NDEBUG
+#endif
         return reportError(report, error, QString(), parent);
     }
 
@@ -142,7 +149,37 @@ KCModule *KCModuleLoader::loadModule(const KCModuleInfo &mod, ErrorReporting rep
      */
     return reportError(report,
                        i18n("The module %1 is not a valid configuration module.", mod.moduleName()),
-                       i18n("<qt>The diagnosis is:<br />The desktop file %1 does not specify a library.</qt>", mod.fileName()), parent);
+                       i18n("<qt>The diagnosis is:<br />The desktop file %1 does not specify a library.</qt>", mod.fileName()),
+                       parent);
+}
+
+KCModule *KCModuleLoader::loadModule(const KPluginMetaData &metaData, QWidget *parent, const QVariantList &args)
+{
+    Q_ASSERT(metaData.isValid());
+    if (!KAuthorized::authorizeControlModule(metaData.pluginId())) {
+        return reportError(ErrorReporting::Inline,
+                           i18n("The module %1 is disabled.", metaData.pluginId()),
+                           i18n("The module has been disabled by the system administrator."),
+                           parent);
+    }
+    const QVariantList args2 = QVariantList(args) << metaData.rawData().value(QStringLiteral("X-KDE-KCM-Args")).toArray();
+
+    KPluginLoader loader(metaData.fileName());
+    KPluginFactory *factory = loader.factory();
+    if (!factory) {
+        qCDebug(KCMUTILS_LOG) << "Couldn't load plugin" << loader.fileName() << ":" << loader.errorString();
+    } else if (auto kcm = std::unique_ptr<KQuickAddons::ConfigModule>(factory->create<KQuickAddons::ConfigModule>(nullptr, args2))) {
+        if (!kcm->mainUi()) {
+            return reportError(ErrorReporting::Inline, i18n("Error loading QML file."), kcm->errorString(), parent);
+        }
+        return new KCModuleQml(std::move(kcm), parent, args2);
+    } else if (auto kcm = factory->create<KCModule>(parent, args2)) {
+        return kcm;
+    } else {
+        qCWarning(KCMUTILS_LOG) << "Error creating object from plugin" << loader.fileName();
+    }
+
+    return reportError(ErrorReporting::Inline, loader.errorString(), QString(), parent);
 }
 
 void KCModuleLoader::unloadModule(const KCModuleInfo &mod)
@@ -154,40 +191,25 @@ void KCModuleLoader::unloadModule(const KCModuleInfo &mod)
 
 bool KCModuleLoader::isDefaults(const KCModuleInfo &mod, const QStringList &args)
 {
-    if (!mod.service() || mod.service()->noDisplay() || mod.library().isEmpty()) {
-        return true;
-    }
-
-    QVariantList args2(args.cbegin(), args.cend());
-
-    KPluginLoader loader(KPluginLoader::findPlugin(QLatin1String("kcms/") + mod.service()->library()));
-    KPluginFactory* factory = loader.factory();
-    if (factory) {
-        std::unique_ptr<KCModuleData> probe(factory->create<KCModuleData>(nullptr, args2));
-        if (probe) {
-            return probe->isDefaults();
-        }
-    }
-
-    std::unique_ptr<KCModuleData> probe(mod.service()->createInstance<KCModuleData>(nullptr, args2));
-    if (probe) {
-        return probe->isDefaults();
+    std::unique_ptr<KCModuleData> moduleData(loadModuleData(mod, args));
+    if (moduleData) {
+        return moduleData->isDefaults();
     }
 
     return true;
 }
 
-
 KCModule *KCModuleLoader::reportError(ErrorReporting report, const QString &text,
                                       const QString &details, QWidget *parent)
 {
     QString realDetails = details;
     if (realDetails.isNull()) {
-        realDetails = i18n("<qt><p>Possible reasons:<ul><li>An error occurred during your last "
-                           "system upgrade, leaving an orphaned control module behind</li><li>You have old third party "
-                           "modules lying around.</li></ul></p><p>Check these points carefully and try to remove "
-                           "the module mentioned in the error message. If this fails, consider contacting "
-                           "your distributor or packager.</p></qt>");
+        realDetails = i18n(
+            "<qt><p>Possible reasons:<ul><li>An error occurred during your last "
+            "system upgrade, leaving an orphaned control module behind</li><li>You have old third party "
+            "modules lying around.</li></ul></p><p>Check these points carefully and try to remove "
+            "the module mentioned in the error message. If this fails, consider contacting "
+            "your distributor or packager.</p></qt>");
     }
     if (report & KCModuleLoader::Dialog) {
         KMessageBox::detailedError(parent, text, realDetails);
@@ -198,3 +220,27 @@ KCModule *KCModuleLoader::reportError(ErrorReporting report, const QString &text
     return nullptr;
 }
 
+KCModuleData *KCModuleLoader::loadModuleData(const KCModuleInfo &mod, const QStringList &args)
+{
+    if (!mod.service() || mod.service()->noDisplay() || mod.library().isEmpty()) {
+        return nullptr;
+    }
+
+    QVariantList args2(args.cbegin(), args.cend());
+
+    KPluginLoader loader(KPluginLoader::findPlugin(QLatin1String("kcms/") + mod.service()->library()));
+    KPluginFactory *factory = loader.factory();
+    if (factory) {
+        KCModuleData *probe(factory->create<KCModuleData>(nullptr, args2));
+        if (probe) {
+            return probe;
+        }
+    }
+
+    KCModuleData *probe(mod.service()->createInstance<KCModuleData>(nullptr, args2));
+    if (probe) {
+        return probe;
+    }
+
+    return nullptr;
+}
